from pykeedy import VMS
from pykeedy.analysis import conditional_entropy, shannon_entropy
from typing import Literal
from pykeedy.crypt import greshko_decrypt, naibbe_encrypt
from pykeedy.utils import load_corpus, preprocess
import numpy as np


def test_reconstruction(text: str, n: int = 1000) -> float:
    def levenshtein(a, b):
        if len(a) < len(b):
            return levenshtein(b, a)
        if len(b) == 0:
            return len(a)

        prev = list(range(len(b) + 1))
        for i, ca in enumerate(a):
            curr = [i + 1]
            for j, cb in enumerate(b):
                curr.append(
                    min(
                        prev[j + 1] + 1,  # deletion
                        curr[j] + 1,  # insertion
                        prev[j] + (ca != cb),
                    )
                )  # substitution
            prev = curr
        return prev[-1]

    avg = 0
    pre = preprocess(text)
    for i in range(n):
        decoded = greshko_decrypt(
            naibbe_encrypt(text, prngseed=np.random.randint(0, 2**32))
        )
        correct = len(pre) - levenshtein(decoded, pre)
        # print(decoded)
        # print(pre)
        # print(correct)
        # print(len(pre))
        avg += correct / len(pre)

    rec = avg / n
    print(
        f"Reconstruction accuracy: {rec * 100:.2f}% over {n} trials of text length {len(pre)}"
    )
    return rec


def test_entropy(
    encode_seeds: int = 1, mode: Literal["char", "word"] = "char"
) -> dict[str, tuple[float, float]]:
    # Load all available texts and encrypt each encode_seeds times
    # Add vms to corpus then calculate shannon & conditional entropy of each and return dict {name: (shannon, conditional)}
    if mode not in ("char", "word"):
        raise ValueError("mode must be 'char' or 'word'")
    plain = load_corpus()
    all: dict = {}
    for name, text in plain.items():
        all[name] = text
        for i in range(encode_seeds):
            all[name + f"_enc{i}"] = naibbe_encrypt(text, prngseed=i)
    all["vms"] = VMS.to_text()
    if mode == "word":
        for name, text in all.items():
            all[name] = text.split(" ")
    results = {}
    for name, text in all.items():
        results[name] = (shannon_entropy(text), conditional_entropy(text))

    return results



import numpy as np
from collections import Counter

from collections import defaultdict
import re


def frequency_rank(
    text: str | list[str], n: int = 1, normalize: bool = True
) -> dict[str, float]:
    # Accept either a string and return character frequency rank,
    # Or a list of words and return word frequency rank.
    # Return sorted descending
    char_level = isinstance(text, str)

    seq = np.array(list(text)) if char_level else np.array(text)

    if len(seq) < n:
        return {}

    ngrams = []
    exclude = np.array(list(" .\n"))  # exclude ' ', '.' and '\n' chars from all grams
    for i in range(len(seq) - n + 1):
        gram = seq[i : i + n]
        if np.any(np.isin(gram, exclude)):
            continue
        ngrams.append(gram)

    unique, counts = np.unique(ngrams, axis=0, return_counts=True)

    sorted_indices = np.argsort(counts)[::-1]
    joiner = "" if char_level else " "

    result: dict[str, int | float] = {}
    for idx in sorted_indices:
        key = joiner.join(unique[idx])  # works for all n, including n==1
        if " " in key:
            key = f"'{key}'"
        count = int(counts[idx])
        result[key] = count

    if normalize:
        total = sum(result.values())
        for key in result:
            result[key] = result[key] / total

    return result


def cooccurence_matrix(
    text: str | list[str], n: int = 2, normalize: bool = True
) -> tuple[list[str], list[list[float]]]:
    char_level = isinstance(text, str)
    seq = np.array(list(text)) if char_level else np.array(text)

    if len(seq) < n:
        return [], []

    # Generate n-grams, excluding any containing spaces
    # TODO add option that includes spaces to calculate most common first and last letters
    ngrams = []
    exclude = [" ", "\n", "."]
    for i in range(len(seq) - n + 1):
        ngram = tuple(seq[i : i + n])
        if any(elem in exclude for elem in ngram):
            continue
        ngrams.append(ngram)

    # Get unique n-grams and their counts
    unique_ngrams, counts = np.unique(ngrams, axis=0, return_counts=True)

    # Counting the gram occurences in all ngram occurences
    # Note this duplicates anything not at the start or end of a sequence
    # We want that in this case
    element_counts = {}
    for ngram_array, count in zip(unique_ngrams, counts):
        for element in ngram_array:
            element_counts[element] = element_counts.get(element, 0) + count

    # Generate sorted-descending list of elements by their counts
    sorted_elements = sorted(
        element_counts.keys(), key=lambda x: element_counts[x], reverse=True
    )

    # Create element to index mapping
    element_to_idx = {elem: i for i, elem in enumerate(sorted_elements)}

    # Initialize matrix
    matrix = np.zeros(tuple(len(sorted_elements) for _ in range(n)), dtype=float)

    # Fill matrix using unique n-grams and their counts
    # This works for any n
    for i, ngram_array in enumerate(unique_ngrams):
        ngram_count = counts[i]
        # Reverseing elements to make columns first element and rows second
        idx = tuple(element_to_idx[elem] for elem in ngram_array)[::-1]
        matrix[idx] += ngram_count

    if normalize:
        matrix = matrix / np.sum(matrix)

    return sorted_elements, matrix.tolist()


def shannon_entropy(text: str | list[str]) -> float:
    # Accept either a string and return character entropy,
    # Or a list of words and return word entropy.
    seq = np.array(list(text)) if isinstance(text, str) else np.array(text)
    unique_chars, counts = np.unique(seq, return_counts=True)
    probabilities = counts / len(text)
    entropy = -np.sum(probabilities * np.log2(probabilities + 1e-10))
    return float(entropy)


def joint_entropy(text: str | list[str], n: int = 2) -> float:
    seq = np.array(list(text)) if isinstance(text, str) else np.array(text)

    if len(seq) < n:
        return 0.0

    # Get each ngram from sequence
    ngrams = [tuple(seq[i : i + n]) for i in range(len(seq) - n + 1)]
    unique_ngrams, counts = np.unique(ngrams, axis=0, return_counts=True)
    probabilities = counts / len(ngrams)
    entropy = -np.sum(probabilities * np.log2(probabilities + 1e-10))
    return float(entropy)


def conditional_entropy(text: str | list[str], n: int = 2) -> float:
    return joint_entropy(text, n) - shannon_entropy(text)


def length_distribution(words: list[str]) -> tuple[tuple[int, int]]:
    token_length_counts = Counter(len(word) for word in words)
    return tuple(sorted(token_length_counts.items()))  # type: ignore


def all_pos(
    text: str, substring: str, word_mode: bool = True, normalize: bool = True
) -> list[float | int]:
    if not len(substring) or not len(text):
        raise ValueError("text and substring must be non-empty strings")

    # Word mode - must be bounded by one of these chars on both sides
    # Not-word-mode - no boundary requirement, which means that
    # anything longer than 1 could have overlaps (2 'aa' in 'aaa')
    word_bounds = set(".\n")

    indexes = []
    start = 0
    word_len = len(substring)

    while True:
        index = text.find(substring, start)
        if index == -1:
            break
        if word_mode:  # Word mode - check boundaries
            start_valid = index == 0 or text[index - 1] in word_bounds
            end_valid = (
                index + word_len == len(text) or text[index + word_len] in word_bounds
            )

            if start_valid and end_valid:
                indexes.append(index)
        else:  # Letter mode - any match means add it
            indexes.append(index)

        start = index + 1
    if normalize:
        # We want both letters and word positions to be normalized to [0, 1]
        # For letters, they can't be at index len(text), so dividing by it
        # means the max normalized value is somewhat less than 1. so sub 1 to make it exact
        # For words, they can only be at the index after the rightmost .
        # (we're assuming no trailing .'s)
        # Same normalization problem, so set total to 1 to right of rightmost .
        total = text.rfind(".") + 1 if word_mode else len(text) - 1
        if total == 0:
            total = len(text)
        indexes = [idx / total for idx in indexes]
    return indexes


# This can tell you how many characters into a string something is
# But what if I want to know how many words into a line a word is
# Or how many lines into a page
# Or how many pages into the manuscript
# Not sure how much the difference would matter in practice?
def position_distribution(
    items: list[str] | set[str],
    sequences: list[str],
    word_mode: bool = False,
    normalize: bool = True,
    average: bool = False,
) -> dict[str, list[int | float] | int | float]:
    """
    Find all positions of each item in items within the list of sequences.
    """
    items = list(dict.fromkeys(items))  # deduplicate, preserve order
    if not items:
        return {}
    results = defaultdict(list)

    if word_mode:
        # Compile regex to match any of the items as whole words
        # Word boundaries (\b) ensure proper token separation
        # Better when len(items) is large: Aho Corasick algorithm
        pattern = re.compile(r"\b(" + "|".join(map(re.escape, items)) + r")\b")

        for seq in sequences:
            for match in pattern.finditer(seq):
                item = match.group(1)
                pos = match.start()
                if normalize:
                    pos /= len(seq)
                results[item].append(pos)

    else:
        pattern = re.compile("|".join(map(re.escape, items)))

        for seq in sequences:
            for match in pattern.finditer(seq):
                item = match.group(0)
                pos = match.start()
                if normalize:
                    pos /= len(seq)
                results[item].append(pos)

    # Drop items with no matches
    results = {k: v for k, v in results.items() if v}

    if average:
        results = {k: sum(v) / len(v) for k, v in results.items()}
        ordered = {
            k: results[k] for k in items if k in results
        }  # preserve original order
        results = ordered
    else:
        # Sort descending by number of occurrences
        results = dict(sorted(results.items(), key=lambda x: len(x[1]), reverse=True))

    return results  # type: ignore



from pydantic import BaseModel, model_validator
import yaml
from pathlib import Path
import importlib.resources as resources
from functools import lru_cache
from typing import ClassVar


class NaibbeEncoding(BaseModel):
    name: str | None = None
    ngram_slot_tables: list[
        dict[str, list[str]]
    ]  # [unigram, bigram_prefix, bigram_suffix, ...] - order extremely important for lists!
    table_odds: (
        list[float] | list[list[float]]
    )  # e.g. [5, 2, 2, 2, 1, 1] - order corresponding to lists in ngram_slot_tables
    ngram_odds: list[float] | None = (
        None  # [1, 1] for equal odds unigram + bigram - order is ascending (first position is unigram, second bigram)
    )

    VALID_TABLES: ClassVar[tuple] = (
        "unigram",
        "bigram_prefix",
        "bigram_suffix",
        "trigram_prefix",
        "trigram_core",
        "trigram_suffix",
    )

    @model_validator(mode="after")
    def check_encoding(self) -> "NaibbeEncoding":
        if self.ngram_odds is None:
            self.ngram_odds = [1.0]
        num_ngrams = len(self.ngram_odds)
        num_tables = num_ngrams * (num_ngrams + 1) // 2
        # one slot per table
        # unigram only -> 1 table
        # unigram + bigram -> 3 tables (Greshko encoding)
        # unigram + bigram + trigram -> 6 tables

        if len(self.ngram_slot_tables) != num_tables:
            raise ValueError(
                "Number of possible ngram lengths and number of tables do not match"
            )

        # Limit to trigram (most of code should work for higher, limit mostly for more clarity in certain areas)
        if len(self.ngram_odds) > 3:
            raise ValueError("Only up to trigrams are supported")

        alphabet = {}
        num_encodings = 0
        for i, tab in enumerate(self.ngram_slot_tables):
            if i == 0:
                alphabet = set(tab.keys())
                if any(len(char) > 1 for char in alphabet):
                    raise ValueError(
                        "ngram slot tables must have only single letters as keys"
                    )
            else:
                if set(tab.keys()) != alphabet:
                    raise ValueError(
                        "All ngram_tables must use the same plaintext alphabet"
                    )

            lengths = set(len(v) for v in tab.values())
            if len(lengths) != 1:
                raise ValueError(
                    f"Each encoding table must have same number of encodings per character (problem with table: {self.tabname(i)})"
                )

            if i == 0:
                num_encodings = len(list(tab.values())[0])
                if isinstance(
                    self.table_odds[0], float
                ):  # single common odds list for all slots
                    if len(self.table_odds) != num_encodings:
                        raise ValueError(
                            "If table_odds is a single list, it must have the same length as the number of encodings per character"
                        )
                elif isinstance(
                    self.table_odds[0], list
                ):  # leaving the door open for slots having different numbers of encodings in future (len would be checked)
                    if len(self.table_odds[i]) != len(self.ngram_slot_tables):  # type: ignore
                        raise ValueError(
                            "If table_odds is a list of lists, it must have the same length as ngram_tables"
                        )
            else:
                if len(list(tab.values())[0]) != num_encodings:
                    raise ValueError(
                        "All tables must have the same number of encodings per character"
                    )

        return self

    @classmethod
    def from_name(cls, name: str) -> "NaibbeEncoding":
        encodings = load_encodings()
        if name not in encodings:
            raise ValueError(
                f"Encoding '{name}' not found. Available encodings: {list(encodings.keys())}"
            )
        filename = encodings[name]
        enc_dir = resources.files("pykeedy.data.encodings")
        filepath = enc_dir / filename
        with resources.as_file(filepath) as file:
            return cls.from_file(file)

    @classmethod
    def from_file(cls, filepath: Path) -> "NaibbeEncoding":
        with open(filepath, "r") as f:
            data: dict = yaml.safe_load(f)
        try:
            data = data["encoding"]
        except KeyError:
            raise ValueError(
                "Encoding YAML file must contain 'encoding' as top-level key"
            )

        tables = []
        valid_tables: tuple = cls.VALID_TABLES  # type: ignore
        for tabname in valid_tables:
            if tabname in data:
                tables.append(data.pop(tabname))
        if len(tables) == 0:
            raise ValueError(
                f"No valid encoding tables found in file. Valid names: {valid_tables}"
            )
        data["ngram_slot_tables"] = tables

        for valname in ("name", "table_odds", "ngram_odds"):
            if valname not in data:
                raise ValueError(f"Encoding YAML file must contain '{valname}' key")

        return cls(**data)

    def to_file(self, filepath: Path) -> None:
        if self.name is None:
            raise ValueError("Encoding must have a name to be saved to file")
        data = {
            "name": self.name,
            "table_odds": self.table_odds,
            "ngram_odds": self.ngram_odds,
        }
        valid_tables = (
            "unigram",
            "bigram_prefix",
            "bigram_suffix",
            "trigram_prefix",
            "trigram_core",
            "trigram_suffix",
        )
        for i, tabname in enumerate(valid_tables):
            if i < len(self.ngram_slot_tables):
                data[tabname] = self.ngram_slot_tables[i]

        with open(filepath, "w") as f:
            yaml.dump({"encoding": data}, f)

    def print(self) -> None:
        print("NaibbeEncoding(")
        print("    ngram_slot_tables = [")

        for i, table in enumerate(self.ngram_slot_tables):
            print(f"        # {self.tabname(i)} table")
            print("        {")
            for key, value in table.items():
                print(f'            "{key}": {value},')
            print("        },")

        print("    ],")
        print(f"    table_odds = {self.table_odds},")
        print(f"    ngram_odds = {self.ngram_odds}")
        print(")")

    def save(self) -> None:
        if self.name is None:
            raise ValueError("Encoding must have a name to be saved to file")
        self.to_file(Path(f"{self.name}.yaml"))

    @property
    def alphabet(self) -> str:
        return "".join(sorted(self.ngram_slot_tables[0].keys()))

    @property
    def common_table_odds(self) -> bool:
        # Checks if table_odds is a single list (same odds for all slots) or a list of lists (different odds for each slot)
        return isinstance(self.table_odds[0], float)

    @property
    def get_slot_decrypt_tables(self) -> list[dict[str, str]]:
        # return: [slot1decode, slot2decode, slot3decode]
        # slot1decode: {"qokchedy": "h", "okchedy": "h", ...}
        return [
            {enc: char for char, encs in slot_dict.items() for enc in encs}
            for slot_dict in self.ngram_slot_tables
        ]

    @property
    def get_slot_lists(self) -> list[list[str]]:
        # return: [slot1list, slot2list, slot3list]
        # slot1list: ["qokchedy", "okchedy", ...]
        return [list(dec.keys()) for dec in self.get_slot_decrypt_tables]  # type: ignore

    def tabname(self, tab: int) -> str:
        names = (
            "unigram",
            "bigram prefix",
            "bigram suffix",
            "trigram prefix",
            "trigram core",
            "trigram suffix",
        )
        try:
            return names[tab]
        except IndexError:
            raise ValueError("Only up to trigrams are supported")

    def ambiguousity(self):
        # Debug for now, later return a value for exactly how ambiguous encoding is.

        # An encoding is ambiguous if and only if:
        # -For any slot (unigram, bigram prefix, bigram suffix) there is a duplicate (among any character or table)
        # -An encoding from an earlier slot can be built from a combination of later ones
        # -There is overlap in the valid ngrams for n>2

        def find_duplicates(lst):
            seen = set()
            duplicates = set()

            for item in lst:
                if item in seen:
                    duplicates.add(item)
                else:
                    seen.add(item)

            return list(duplicates)

        # Check first case
        for i, chartab in enumerate(self.ngram_slot_tables):
            # Encodings within the same slot and letter don't have to be unique to be unambiguous, so take their set
            all_encs = [enc for encs in chartab.values() for enc in set(encs)]
            encset = set(all_encs)
            print(f"{self.tabname(i)}: {len(encset)} / {len(all_encs)} (unique/total)")
            # if len(encset) != len(all_encs):
            #     print(f"Duplicate encodings in {self.tabname(i)} table ({find_duplicates(all_encs)})")
            #     return True

        # Check second case
        if len(self.ngram_odds) == 2:
            unigram_encs = set(
                enc for encs in self.ngram_slot_tables[0].values() for enc in encs
            )
            bigram_prefix_encs = set(
                enc for encs in self.ngram_slot_tables[1].values() for enc in encs
            )
            bigram_suffix_encs = set(
                enc for encs in self.ngram_slot_tables[2].values() for enc in encs
            )

            bigrams = []

            for bpre in bigram_prefix_encs:
                for bsuf in bigram_suffix_encs:
                    bigrams.append(bpre + bsuf)
            bigram_set = set(bigrams)
            print(f"bigrams: {len(bigram_set)} / {len(bigrams)} (unique/total)")
            intersection = bigram_set.intersection(unigram_encs)
            if len(intersection) > 0:
                print(f"Unigram-bigram collision: {len(intersection)}")
                return True
            if len(bigram_set) != len(bigrams):
                print("Intersection within bigrams")
                return True
            return False

        elif len(self.ngram_odds) == 3:
            raise NotImplementedError

        else:
            raise ValueError("Only up to trigram supported")


@lru_cache(maxsize=1)
def get_default_encoding() -> NaibbeEncoding:
    DEFAULT_ENCODING_NAME: str = "greshko_202507"
    return NaibbeEncoding.from_name(DEFAULT_ENCODING_NAME)


def parse_encoding(encoding: NaibbeEncoding | str | None) -> NaibbeEncoding:
    if not isinstance(encoding, NaibbeEncoding):
        if isinstance(encoding, str):
            encoding = NaibbeEncoding.from_name(encoding)
        elif encoding is None:
            encoding = get_default_encoding()
    return encoding


@lru_cache(maxsize=1)
def _load_encodings() -> dict[str, str]:  # [enc_name: filename]
    enc_dir = resources.files("pykeedy.data.encodings")
    result = {}
    for entry in enc_dir.iterdir():
        if entry.is_file() and entry.name.endswith(".yaml"):
            with entry.open("r", encoding="utf-8") as f:
                data = yaml.safe_load(f)

            try:
                result[data["encoding"]["name"]] = entry.name
            except KeyError:
                print(
                    f'Warning: Encoding file {entry.name} does not contain "encoding"->"name" key and cannot be parsed, skipping'
                )
    return result


def load_encodings(force_update: bool = False) -> dict[str, str]:
    if force_update:
        _load_encodings.cache_clear()
    return _load_encodings()



from importlib import resources
import numpy as np
import matplotlib.pyplot as plt
from typing import Sequence
import regex
import re
from dataclasses import dataclass
from pathlib import Path


def preprocess(text: str) -> str:
    # preprocess text by deleting spaces and punctuation then lowercasing
    # \p{C} - Control characters (includes things like null bytes, tabs, newlines)
    # \p{M} - Mark characters (combining marks, accents, diacritics)
    # \p{P} - Punctuation characters (periods, commas, quotes, etc.)
    # \p{S} - Symbol characters (mathematical symbols, currency symbols, etc.)
    # \p{Z} - Separator characters including spaces, tabs, etc
    # regex is imported because standard re does not this type of syntax
    remove = regex.compile(
        r"[\p{P}|\p{S}]+", regex.UNICODE
    )  # space at end of pattern is important!
    text = remove.sub("", text)
    text = re.sub(r"(?:\r?\n){2,}", "\n", text)
    text = text.lower()
    return text


@dataclass
class PlainManuscript:
    text: str

    def to_text(self) -> str:
        return self.text

    def to_words(self) -> list[str]:
        return self.text.split()

    def to_lines(self) -> list[str]:
        return self.text.splitlines()


def load_corpus(
    from_dir: str | None = None,
    names: str | list[str] | None = None,
    prep: bool = True,
    limit_length: int | None = 250_000,
    give_objects: bool = False,
) -> dict[str, str | PlainManuscript]:
    """
    Load plaintext(s) and return dict[name: text]
    Loads from library corpus if from_dir is None
    Otherwise uses all .txt files in from_dir, filename (sans .txt) becoming the name
    Accepts single name, list of names, or None for all available texts
    """
    if from_dir is not None:
        # Convert string path to Traversable using pathlib.Path
        text_dir = Path(from_dir)
    else:
        text_dir = resources.files("pykeedy.data.plaintexts")
    result = {}
    for entry in text_dir.iterdir():
        if entry.is_file() and entry.name.endswith(".txt"):
            name = entry.name.strip(".txt")
            if names is not None:
                if isinstance(names, str):
                    if name != names.strip(".txt"):
                        continue
                elif isinstance(names, list):
                    if name not in [n.strip(".txt") for n in names]:
                        continue
            with entry.open("r", encoding="utf-8") as f:
                text = f.read()
            if prep:
                text = preprocess(text)
            if limit_length and len(text) > limit_length:
                orig_len = len(text)
                text = text[:limit_length]
                print(
                    f"Warning: Truncated text '{name}' from {orig_len} to {limit_length} characters"
                )
            result[name] = text
    if not result:
        raise ValueError("No texts found - check from_dir and names arguments")
    if give_objects:
        result = {name: PlainManuscript(text) for name, text in result.items()}

    return result  # type: ignore


def add_axlabels(key: Sequence[str]) -> None:
    if not all(isinstance(x, str) for x in key) or len(key) != 2:
        raise ValueError("key must contain only 2 strings representing labels")

    plt.xlabel(key[0])
    plt.ylabel(key[1])


def scatterplot(
    d: dict, key: Sequence[str] = ("X", "Y"), fname: str = "scatterplot.png"
) -> None:
    # Expect dict of {name: (x, y)} pairs
    # name: str, x & y: float
    for i, (name, (x, y)) in enumerate(d.items()):
        plt.scatter(x, y, label=name, c=f"C{i}")
        plt.annotate(name, (x, y), xytext=(5, 5), textcoords="offset points")
    add_axlabels(key)
    plt.savefig(fname)
    plt.close()
    print(f"Saved scatter plot to {fname}")


def barplot(
    d: dict,
    key: Sequence[str] = ("Item", "Number"),
    fname: str = "barplot.png",
    n_max: int = 20,
    color: str | None = None,
    title: str | None = None,
) -> None:
    names, values = zip(*d.items())
    names, values = list(names), list(values)
    if len(names) > n_max:
        print(f"Warning: More than {n_max=} items passed to barplot, truncating")
        names, values = names[:n_max], values[:n_max]
    plt.bar(names, values, color=color)
    add_axlabels(key)
    if title:
        plt.title(title)
    plt.xticks(rotation=45, ha="right")
    plt.tight_layout()
    plt.savefig(fname)
    plt.close()
    print(f"Saved bar plot to {fname}")


def heatmap(
    labels: list[str],
    matrix: list[list[float]],
    key: Sequence[str] = ("First element", "Second element"),
    fname: str = "heatmap.png",
    n_max: int = 20,
    title: str | None = None,
) -> None:
    # Convert to numpy array for easier handling
    matrix_array = np.array(matrix)

    if len(labels) > n_max:
        print(
            f"Warning: More than {n_max=} items passed to heatmap, truncating labels and matrix"
        )
        labels = labels[:n_max]
        matrix_array = matrix_array[:n_max, :n_max]

    # Create the heatmap
    fig, ax = plt.subplots(figsize=(8, 6))
    im = ax.imshow(matrix_array, cmap="plasma", aspect="auto")
    if title:
        plt.title(title, pad=15)

    # Set the ticks and labels
    ax.set_xticks(range(len(labels)))
    ax.set_yticks(range(len(labels)))
    ax.set_xticklabels(labels)
    ax.set_yticklabels(labels)

    ax.xaxis.tick_top()
    plt.xticks(rotation=45, ha="left")
    ax.xaxis.set_label_position("top")

    # Add colorbar
    plt.colorbar(im)

    # Add text annotations showing the values
    if len(labels) < 10:
        for i in range(len(labels)):
            for j in range(len(labels)):
                ax.text(
                    j, i, matrix_array[i, j], ha="center", va="center", color="black"
                )

    add_axlabels(key)

    plt.tight_layout()
    plt.savefig(fname)
    plt.close()
    print(f"Saved heatmap to {fname}")


def seriesplot(
    d: dict,
    key: Sequence[str] = ("X", "Y"),
    fname: str = "seriesplot.png",
    title: str | None = None,
) -> None:
    # Expect dict of {name: ((x1,y1), (x2,y2), ...)} pairs
    # name: str, x & y: float
    for i, (name, points) in enumerate(d.items()):
        x_vals, y_vals = zip(*points)
        plt.plot(
            x_vals, y_vals, label=name, c=f"C{i}", marker="o", linewidth=2, markersize=4
        )

    if title:
        plt.title(title)

    plt.legend()

    add_axlabels(key)

    plt.tight_layout()
    plt.savefig(fname)
    plt.close()
    print(f"Saved series plot to {fname}")



import json
import os


def analysis_to_md(
    manifest_path: str = "manifest.json",
    output_markdown_fname: str = "analysis_report.md",
    md_to_imgs_path: str = "./",
    total_width: int = 1800,
) -> None:
    """
    Generate a markdown file from a manifest.json that lists analysis images.

    Args:
        manifest_path: Path to the manifest.json file
        output_markdown_path: Path where the markdown file will be written
        image_path: Path prefix for accessing images in the markdown
        total_width: Total width to distribute among per-manuscript images

    Raises:
        FileNotFoundError: If manifest.json doesn't exist
        ValueError: If required files are missing
        KeyError: If manifest structure is invalid
    """

    # Check for and load manifest.json
    if not os.path.exists(manifest_path):
        raise FileNotFoundError(f"Manifest file not found: {manifest_path}")

    with open(manifest_path, "r") as f:
        manifest = json.load(f)

    # Extract manuscript names and analysis structure
    try:
        manuscripts = manifest["analysis"]["manuscripts"]
        cross_manuscript = manifest["analysis"]["cross_manuscript"]
        per_manuscript = manifest["analysis"]["per_manuscript"]
    except KeyError as e:
        raise KeyError(f"Invalid manifest structure: missing key {e}")

    # Generate all expected PNG filenames
    expected_files = []

    # Add cross-manuscript files
    expected_files.extend(cross_manuscript)

    # Add per-manuscript files (substitute <name> with each manuscript name)
    for analysis_type, pattern_list in per_manuscript.items():
        for pattern in pattern_list:
            for manuscript in manuscripts:
                filename = pattern.replace("<name>", manuscript)
                expected_files.append(filename)

    # Check that all files exist
    missing_files = []
    for filename in expected_files:
        # Remove the image_path prefix if it's already in the filename for checking
        file_to_check = filename
        if not filename.startswith(md_to_imgs_path):
            file_to_check = os.path.join(os.path.dirname(manifest_path), filename)

        if not os.path.exists(file_to_check):
            missing_files.append(filename)

    if missing_files:
        raise ValueError(f"Missing files: {', '.join(missing_files)}")

    # Create linear list of filename patterns for markdown generation
    markdown_patterns = []

    # Add cross-manuscript patterns
    markdown_patterns.extend(cross_manuscript)

    # Add per-manuscript patterns (keeping <name> placeholder)
    for analysis_type, pattern_list in per_manuscript.items():
        markdown_patterns.extend(pattern_list)

    # Generate markdown content
    markdown_lines = []

    breaks = 5

    # Process cross-manuscript images first
    for pattern in cross_manuscript:
        image_src = os.path.join(md_to_imgs_path, pattern).replace("\\", "/")
        markdown_lines.append('<p float="left">')
        markdown_lines.append(f'  <img src="{image_src}" width="{total_width}" />')
        markdown_lines.append("</p>")
    markdown_lines.append("<br></br>\n" * breaks)
    # Process per-manuscript images grouped by type
    num_manuscripts = len(manuscripts)
    individual_width = (
        total_width // num_manuscripts if num_manuscripts > 0 else total_width
    )

    for analysis_type, pattern_list in per_manuscript.items():
        for pattern in pattern_list:
            markdown_lines.append('<p float="left">')

            for manuscript in manuscripts:
                filename = pattern.replace("<name>", manuscript)
                image_src = os.path.join(md_to_imgs_path, filename).replace("\\", "/")
                markdown_lines.append(
                    f'  <img src="{image_src}" width="{individual_width}" />'
                )

            markdown_lines.append("</p>")
            markdown_lines.append("<br></br>\n" * breaks)

    # Write markdown file
    with open(output_markdown_fname, "w") as f:
        f.write("\n".join(markdown_lines))

    print(f"Markdown file generated: {output_markdown_fname}")



from pykeedy.naibbe import NaibbeEncoding, parse_encoding
import numpy as np
from pykeedy.utils import preprocess
import re


def naibbe_encrypt(
    text: str, encoding: NaibbeEncoding | str | None = None, prngseed: int | None = 42
) -> str:
    """
    Encrypt text using some Naibbe encoding.
    Note that this is a general implementation for encodings with any number of characters, ngram lengths, probabilities, etc.
    A function for only unigram-bigram encodings would be simpler.
    """
    encoding = parse_encoding(encoding)
    text = preprocess(text)
    alphabet_exclude = f"[^{encoding.alphabet}]"
    orig_len = len(text)
    text = re.sub(alphabet_exclude, "", text)
    if len(text) != orig_len:
        print(
            f"Warning: {orig_len - len(text)}/{orig_len} characters were removed from input text because they are not in the encoding alphabet"
        )

    if prngseed is None:
        prngseed = np.random.randint(0, 2**32)
    rng = np.random.default_rng(seed=prngseed)

    def odds_to_thresholds(odds: list[float]) -> np.ndarray:
        # process ordered odds lists to make it easy to select option (ngram length, or, table to use) from a randomly generated number
        # normalize sum of list to 1, then cumsum to get thresholds - last element should always be ~1,
        # which is discarded (rolled to first element then replaced with 0)
        # example: [1 1 1] -> [0 0.3333 0.6666]

        oddsarr = np.array(odds)
        thresharr = np.roll(np.cumsum(oddsarr / np.sum(oddsarr)), 1)
        thresharr[0] = 0.0
        return thresharr

    ngram_length_thresholds = odds_to_thresholds(encoding.ngram_odds)  # type: ignore - not None guaranteed during model validation

    table_thresholds_array = None
    # make and copy single thresholds table if each has the same odds, otherwise do for each
    # copying means we just assume each table has its own odds instead of branching later
    if encoding.common_table_odds:
        table_thresholds = odds_to_thresholds(encoding.table_odds)  # type: ignore
        table_thresholds_array = np.array(
            [table_thresholds for _ in range(len(encoding.ngram_slot_tables))]
        )
    else:
        table_thresholds_array = np.array(
            [odds_to_thresholds(to) for to in encoding.table_odds]
        )  # type: ignore

    def select_option(thresholds: np.ndarray) -> int:
        # using the precomputed threshold table, generate a random number, then see where it falls on the table (the largest index it is larger than)
        # to select how many characters to encode in a word or which table to use for a character
        rand = rng.random()  # 0 -> 1
        return np.nonzero(thresholds < rand)[0][-1]

    i = 0
    encoded = ""
    while i < len(text):
        gramsize = select_option(ngram_length_thresholds) + 1  # select ngram size

        if i + gramsize > len(
            text
        ):  # not enough characters left for selected ngram size, so use smaller
            gramsize = len(text) - i

        start_table_idx = (
            0 if gramsize == 1 else 1 if gramsize == 2 else 3
        )  # Only support up to trigrams (same as NaibbeEncoding)

        word = ""

        for j in range(gramsize):
            char = text[i + j]
            if char not in str(encoding.alphabet):
                raise ValueError(
                    f"Character '{char}' at position {i + j} not in encoding alphabet"
                )
            table = select_option(table_thresholds_array[start_table_idx + j])
            word += encoding.ngram_slot_tables[start_table_idx + j][char][table]
        i += gramsize
        encoded += word + " "

    return encoded.strip()


def greshko_decrypt(encoded: str, encoding: NaibbeEncoding | str | None = None) -> str:
    # IMPORTANT: If the encoding is ambiguous (Greshko encoding is), some information is lost in the encryption/decryption process. With this decoding algorithm and the current encoding algorithm reconstruction rate is 95%.

    # This uses the algorithm from the paper (Greshko 2025). It makes use of properties of the encoding tables so it will not work for other encodings!
    # Strategy is important because the encoding is not just a little ambiguous - 90% of unigrams are also valid bigrams.
    # Step 1: If word is a valid unigram, decode it as such (it is usually the far more likely option even if it's also a bigram)
    # Step 2: Use "breakpoint" strings: There are strings unique to one of the two word types that bigrams are built from (type 1/2 affixes). Type 1 is very likely to be a prefix and type 2 to be a suffix, so the rightmost breakpoint string in a suffix or leftmost in a prefix defines the breakpoint, both affixes are then parsed (prefix and suffix tables are unique)
    # Step 3: Refer to a grammar slot table. Check each character against successive slots until there's a match, then decode the prefix as the longest valid string.

    # One can imagine algorithms that do various types of "greedy parsing" to find the likeliest decoding instead of using a hardcoded method and this is a future goal.
    # When doing this it will also be important to focus on the 15th-century-human-doability as the paper does.

    encoding = parse_encoding(encoding)

    if encoding.name != "greshko_202507":
        raise NotImplementedError("Decryption for other encodings not yet implemented")

    prefix_only_strs = ["ch", "sh", "cfh", "ckh", "cph", "cth", "f", "k", "p", "t", "x"]
    suffix_only_strs = ["a", "e", "g", "i", "m", "n"]

    t1_slots = [
        ["q", "s", "d", "x"],
        ["o", "y"],
        ["d", "r"],
        ["t", "k", "p", "f"],
        ["ch", "sh"],
        ["cth", "ckh", "cph", "cfh"],
    ]

    t2_slots = [
        ["e", "ee", "eee", "g"],
        ["s", "d"],
        ["o", "a"],
        ["i", "ii", "iii"],
        ["d", "l", "r", "m", "n"],
        ["s"],
        ["y"],
    ]

    slot_decrypt_tables: list[dict[str, str]] = encoding.get_slot_decrypt_tables  # type: ignore

    slot_lists: list[list[str]] = encoding.get_slot_lists  # type: ignore

    def common_prefix_length(s1: str, s2: str) -> int:
        # Finds length of common prefix between two strings
        # "Prefix" used here entirely unrelated to voynich grammar
        i = 0
        min_len = min(len(s1), len(s2))

        while i < min_len and s1[i] == s2[i]:
            i += 1

        return i

    def slot_hit(glyph: str, vord_remaining: str) -> bool:
        return common_prefix_length(glyph, vord_remaining) == len(glyph)

    def get_longest_affix(vord: str, slots: list[list[str]]) -> int:
        i = 0
        for slot_options in slots:
            best = 0
            for glyph in slot_options:
                if slot_hit(glyph, vord[i:]):
                    best = len(glyph)
            i += best
        return i

    def parse_from_breakpoint(vord: str, pt: int) -> str:
        pre = vord[:pt]
        suf = vord[pt:]
        return slot_decrypt_tables[1][pre] + slot_decrypt_tables[2][suf]

    def step1(vord: str) -> str | None:
        if vord in slot_lists[0]:
            return slot_decrypt_tables[0][vord]

    def step2(vord: str) -> str | None:
        best = None
        for glyph in prefix_only_strs:  # Get rightmost prefix glyph
            pos = vord.rfind(glyph)
            if pos != -1:
                if best is None or pos > best:
                    best = pos + len(glyph)
        for glyph in suffix_only_strs:  # Get leftmost suffix glyph
            pos = vord.find(glyph)
            if pos != -1:
                if best is None or pos < best:
                    best = pos
        if best and best > 0:
            try:
                # Sometimes the process is just wrong because type 1 affixes can be suffixes and vice versa,
                # so have to catch it and continue to step. words caught here tend to have no type 1 affix glyphs
                # example: daleor, lsdaiin, oldal, alaiin, aleedal, qodain...
                return parse_from_breakpoint(vord, best)
            except KeyError:
                return None

    def step3(vord: str) -> str | None:
        t1 = get_longest_affix(vord, t1_slots)
        t2 = get_longest_affix(vord, t2_slots)
        longest_idx = max(t1, t2)

        try:
            return parse_from_breakpoint(vord, longest_idx)
        except (IndexError, KeyError):
            return None

    vords = encoded.split(" ")
    decoded = ""  # We add to this with each decode
    for vord in vords:  # Each word is entirely independent
        for i, step in enumerate([step1, step2, step3]):
            res = step(vord)  # Result is either successfully decoded text or None
            if res is not None:
                decoded += res
                break
            if i == 2:
                # We're after the last step (i==2) and found nothing
                # (any success would've broken out of the lower loop)
                decoded += "?"

        # Problem grams:
        # qol + o - grammar suggests qo + lo, lo is invalid (same for qolor)
        # o + ry / or + y - is also a valid t2 gram, so greedy parsing leaves no suffix
        # aiir + y - valid t2 gram
    return decoded



# type: ignore
from pykeedy.utils import load_corpus, scatterplot, barplot, seriesplot, heatmap
from pykeedy.analysis import (
    shannon_entropy,
    conditional_entropy,
    frequency_rank,
    cooccurence_matrix,
    length_distribution,
    position_distribution,
)
from pykeedy.crypt import naibbe_encrypt
from pykeedy.datastructures import Manuscript
from pykeedy.utils import PlainManuscript
from datetime import datetime
import time
import os
import json

# This file combines all the real analyses from the examples,
# performs all of them on all of the manuscripts in the library corpus, including VMS,
# and saves all the results as pngs in the current folder,
# along with a manifest.json listing all generated files


# All analyses comparing manuscripts against each other
def cross_manuscript(
    texts: dict[str], output_dir: str, add_encrypted: bool, run_entropy: bool
) -> None:
    all = {}
    # Generate all comparison texts
    for name, ms in texts.items():
        text = ms.to_text()
        all[name] = text
        if add_encrypted:
            # Even do VMS itself as a check on encryption effect
            all[name + "_naibbe"] = naibbe_encrypt(text)

    if run_entropy:
        entropy = {}
        for name, text in all.items():
            entropy[name] = (shannon_entropy(text), conditional_entropy(text))

        # Generate entropy plot
        scatterplot(
            entropy,
            key=("character entropy (bits)", "conditional entropy (bits)"),
            fname=f"{output_dir}/entropy_comparison.png",
        )


# All analyses on individual manuscripts
# Called once per manuscript, so accepts a single one, not multiple like cross_manuscript
def per_manuscript(
    name: str,
    i: int,
    ms: Manuscript | PlainManuscript,
    output_dir: str,
    run_ngrams: bool,
    ngram_max_n: int,
    run_heatmaps: bool,
    run_lengths: bool,
    run_positions: bool,
) -> None:
    # Per-manuscript analysis
    barplot_colors = [
        "red",
        "blue",
        "green",
        "orange",
        "purple",
        "brown",
        "pink",
        "gray",
        "olive",
        "cyan",
    ]
    color = barplot_colors[i]

    # For ngrams and heatmaps, always do them at both character and word level
    for mode in ["char", "word"]:
        if run_ngrams:
            # Generate ngram plots
            text = ms.to_text() if mode == "char" else ms.to_words()
            for i in range(ngram_max_n):
                gramsize = i + 1
                barplot(
                    frequency_rank(text, n=gramsize),
                    key=(f"{mode} {gramsize}-gram", "Frequency"),
                    fname=f"{output_dir}/{name}_{mode}_{gramsize}-gram_freq.png",
                    color=color,
                    title=f"{name} ngram freq",
                )
        if run_heatmaps:
            # Generate heatmaps
            res = cooccurence_matrix(text, n=2)
            heatmap(
                *res,
                fname=f"{output_dir}/{name}_{mode}_cooccurence_heatmap.png",
                title=f"{name} {mode} co-occurence",
            )

    if run_lengths:
        # Generate word length plots
        words = ms.to_words()
        results = {
            "token": length_distribution(words),
            "type": length_distribution(list(set(words))),
        }

        seriesplot(
            results,
            key=("Word length", "Count"),
            fname=f"{output_dir}/{name}_word_lengths.png",
            title=f"{name} word length",
        )

    if run_positions:
        text = ms.to_text()
        letters = set(text)
        word_tokens = ms.to_words()
        lines = ms.to_lines()

        # Positions of letters in words
        all_pos = position_distribution(
            letters, word_tokens, normalize=True, average=False
        )
        many_occurs = {k: sum(v) / len(v) for k, v in all_pos.items() if len(v) > 200}
        barplot(
            many_occurs,
            key=("Letter (most common on left)", "Avg pos in word"),
            fname=f"{output_dir}/{name}_letters_pos_in_words.png",
            n_max=len(many_occurs),
            color=color,
            title=f"{name} letter positions in words",
        )

        # Positions of 20 most common words
        top_words = [k for k, v in frequency_rank(word_tokens).items()][:20]

        barplot(
            position_distribution(top_words, lines, word_mode=True, average=True),
            key=("Word (most common on left)", "Avg pos in line"),
            fname=f"{output_dir}/{name}_words_pos_in_lines.png",
            color=color,
            title=f"{name} word positions in lines",
        )
        barplot(
            position_distribution(top_words, [text], word_mode=True, average=True),
            key=("Word (most common on left)", "Avg pos in VMS"),
            fname=f"{output_dir}/{name}_words_pos_in_manuscript.png",
            color=color,
            title=f"{name} word positions in manuscript",
        )


def run_full_analysis(
    voynich: Manuscript,
    comparison_texts_dir: str | None = None,
    comparison_text_names: list[str] | None = None,
    output_dir: str = "full_analysis_results",
    add_encrypted: bool = True,
    run_entropy: bool = True,
    run_ngram_freqs: bool = True,
    ngram_max_n: int = 3,
    run_cooccurence: bool = True,
    run_word_lengths: bool = True,
    run_positions: bool = True,
) -> None:
    # Currently available analyses
    run_cross_manuscript = any([run_entropy])
    run_per_manuscript = any(
        [run_ngram_freqs, run_cooccurence, run_word_lengths, run_positions]
    )

    start = time.time()

    os.makedirs(output_dir, exist_ok=True)

    if comparison_texts_dir:
        texts = load_corpus(
            from_dir=comparison_texts_dir,
            names=comparison_text_names,
            give_objects=True,
        )
    else:
        texts = load_corpus(names=comparison_text_names, give_objects=True)
    texts["VMS"] = voynich

    if run_cross_manuscript:
        cross_manuscript(
            texts=texts,
            output_dir=output_dir,
            add_encrypted=add_encrypted,
            run_entropy=run_entropy,
        )

    if run_per_manuscript:
        for i, (name, ms) in enumerate(texts.items()):
            per_manuscript(
                name=name,
                i=i,
                ms=ms,
                output_dir=output_dir,
                run_ngrams=run_ngram_freqs,
                ngram_max_n=ngram_max_n,
                run_heatmaps=run_cooccurence,
                run_lengths=run_word_lengths,
                run_positions=run_positions,
            )

    analysis_manifest = {
        "analysis": {
            "generated_at": datetime.now().isoformat(),
            "manuscripts": list(texts.keys()),
            "cross_manuscript": ["entropy_comparison.png"],
            "per_manuscript": {
                # If any of the calculated plots are missing from here they will not be displayed in whatever parses this
                "char_ngram_freq": [
                    "<name>_char_1-gram_freq.png",
                    "<name>_char_2-gram_freq.png",
                    "<name>_char_3-gram_freq.png",
                ],
                "word_ngram_freq": [
                    "<name>_word_1-gram_freq.png",
                    "<name>_word_2-gram_freq.png",
                    "<name>_word_3-gram_freq.png",
                ],
                "cooccurence_heatmaps": [
                    "<name>_char_cooccurence_heatmap.png",
                    "<name>_word_cooccurence_heatmap.png",
                ],
                "word_lengths": ["<name>_word_lengths.png"],
                "positions": [
                    "<name>_letters_pos_in_words.png",
                    "<name>_words_pos_in_lines.png",
                    "<name>_words_pos_in_manuscript.png",
                ],
            },
        }
    }
    manifest_path = os.path.join(output_dir, "manifest.json")
    with open(manifest_path, "w", encoding="utf-8") as f:
        json.dump(analysis_manifest, f, indent=4)

    finish = time.time()
    print(
        f"Completed analysis of {len(texts)} manuscripts in {finish - start:.2f} seconds"
    )



import re
from enum import Enum
from dataclasses import dataclass, fields
from typing import Callable, Sequence, Literal
import importlib.resources as resources
from functools import lru_cache


class LocusPropType:
    # A way to tell that something is a filterable property
    pass


class LocusProp:
    class Location(LocusPropType, Enum):
        UnrelatedToPrev = "@"
        BelowPrev = "+"
        BelowAndLeftOfPrev = "*"
        RightOfPrev = "="
        AlongArcFromPrev = "&"
        RightishFromPrev = "~"
        AboveAndRightOfPrev = "/"
        DoesNotExist = "!"

    class Type(LocusPropType, Enum):
        # P: Linear text in paragraphs
        ParagraphLeftJustified = "P0"
        ParagraphNotLeftJustified = "P1"
        ParagraphFreeFloating = "Pb"
        ParagraphCentered = "Pc"
        ParagraphRightJustified = "Pr"
        ParagraphRightJustifiedTitle = "Pt"

        # L: Short piece of text, a word, or a character that is anywhere on the page, mostly labels
        NotDrawingAssociated = "L0"
        AstronomicalLabel = "La"
        PharmaceuticalContainerLabel = "Lc"
        HerbalFragmentLabel = "Lf"
        BiologicalNymphLabel = "Ln"
        HerbalLargePlantLabel = "Lp"
        StarLabel = "Ls"
        BiologicalTubeLabel = "Lt"
        ExtraneousWriting = "Lx"
        ZodiacLabel = "Lz"

        # C: Text along circumference of circle
        CircumferentialClockwise = "Cc"
        CircumferentialCounterClockwise = "Ca"

        # R: Text along radius of circle
        RadialInwards = "Ri"
        RadialOutwards = "Ro"

    class IllustrationType(LocusPropType, Enum):
        Astronomical = "A"
        Biological = "B"
        Cosmological = "C"
        Herbal = "H"
        Pharmaceutical = "P"
        MarginalStarsOnly = "S"
        TextOnly = "T"
        Zodiac = "Z"

    class CurrierLanguage(LocusPropType, Enum):
        A = "A"
        B = "B"

    class DavisHand(LocusPropType, Enum):
        H1 = "1"
        H2 = "2"
        H3 = "3"
        H4 = "4"
        H5 = "5"
        At = "@"

    class CurrierHand(LocusPropType, Enum):
        C1 = "1"
        C2 = "2"
        C3 = "3"
        C4 = "4"
        C5 = "5"
        X = "X"
        Y = "Y"  # IVTFF format pdf and transliteration do not match - pdf claims X + Z, actual transliteration has X + Y

    class ExtraneousWriting(LocusPropType, Enum):
        ColorAnnotation = "C"
        MonthName = "M"
        Other = "O"
        CharOrNumSequence = "S"
        Various = "V"  # This is deprecated but still in the source for some reason.

    @classmethod
    def prop_types(cls) -> list[type]:
        return [
            LocusProp.Location,
            LocusProp.Type,
            LocusProp.IllustrationType,
            LocusProp.CurrierLanguage,
            LocusProp.DavisHand,
            LocusProp.CurrierHand,
            LocusProp.ExtraneousWriting,
        ]

    @classmethod
    def print_props(cls) -> None:
        for proptype in cls.prop_types():
            print(f"LocusProp.{proptype.__name__}:")
            for item in proptype:  # type: ignore
                print(f"  {item.name}")


class VMSDataclass:
    @classmethod
    def print_fields(cls) -> None:
        print(f"{cls.__name__} fields:")
        for field in fields(cls):  # type: ignore
            print(f"Name: {field.name}")


# Transliteration issues:
#   - Rosette page lacking folio number, inconsistent with other headers (this is the only reason folios_in_quire can be None)
#   - IVTFF source says Currier hands contains X & Z but it actually contains X & Y


@dataclass
class Locus(VMSDataclass):
    # Page properties
    quire_num: int
    page_in_quire_num: int
    folio_in_quire_num: int | None
    bifolio_in_quire_num: int
    illustration: LocusProp.IllustrationType
    currier_language: LocusProp.CurrierLanguage | None
    davis_hand: LocusProp.DavisHand
    currier_hand: LocusProp.CurrierHand | None
    extraneous_writing: LocusProp.ExtraneousWriting | None

    # This can be grabbed from either page or locus level
    # Constructor chooses to grab from locus level
    page_name: str
    # Locus properties

    id_str: str
    locus_in_page_num: int
    location: LocusProp.Location
    type: LocusProp.Type
    text: str

    @classmethod
    def from_line(cls, page_props: tuple, line: str) -> "Locus":
        num_props_from_page_level = 9
        if len(page_props) != num_props_from_page_level:
            raise ValueError(
                f"Incorrect page_props passed to locus constructor, expecting len = {num_props_from_page_level}, got {len(page_props)}"
            )
        # Expects line to be a single locus, with no page headers or comments as lines
        # Example line: "<f89v1.23,@Lf>    opol.olaiin" (no \n)
        # So we want to extract page name, locus number in page, location & type code (3 characters), and text.
        line = re.sub(
            r"<[!@%$].*?>", "", line
        )  # Remove inline comments except interruption tags <[-~]>
        match = re.match(r"<([^.]+)\.([^,]+),([^>]+)>\s*([a-zA-Z?].*)", line)
        # Important parts of above line:

        if not match:
            raise ValueError(f"Line does not match expected locus format: {line}")
        page_name = match.group(1)
        locus_in_page_num = int(match.group(2))

        # Extract 3 character code. First character is location, next two are type.
        loctypecode = match.group(3)
        if len(loctypecode) != 3:  # Must be 3 characters
            raise ValueError(
                f"Corrupted IVTFF file: Locus location & type code must be exactly 3 characters: {loctypecode}"
            )
        locus_location = LocusProp.Location(loctypecode[0])
        locus_type = LocusProp.Type(loctypecode[1:])
        text = match.group(4).strip()
        id_str = f"{page_name}.{locus_in_page_num}"
        return cls(
            *page_props,
            page_name=page_name,
            id_str=id_str,
            locus_in_page_num=locus_in_page_num,
            location=locus_location,
            type=locus_type,
            text=text,
        )

    def props(self) -> list[LocusPropType]:
        return [val for val in self.__dict__.values() if isinstance(val, LocusPropType)]

    def is_label(self) -> bool:
        return "Label" in self.type.name

    def is_paragraph(self) -> bool:
        return "Paragraph" in self.type.name

    def is_below_prev(self) -> bool:
        return "Below" in self.location.name


# Tries to convert strings of form 'letter' or 'numbers' to int.
def to_int(letter: str) -> int:  # A -> 1, B -> 2 ...
    if len(letter) != 1:
        raise ValueError("str must be single character")
    try:
        return int(letter)
    except ValueError:
        if not letter.isalpha():
            raise ValueError("letter str must be alphabetic")
        return ord(letter.upper()) - ord("A") + 1


def loci_list_from_lines(lines: list[str], header: str) -> list[Locus]:
    # Expects each line to be one locus, with no page headers or comments as lines
    # Header example: "$Q=M $P=R $F=y $B=2 $I=B $L=B $H=2 $C=2"
    # We don't care about page name because that is in each line as well as the page header,
    # so the choice was made to extract it at the locus level to ensure it can't be wrong.
    pattern = r"\$([A-Z])=([A-Za-z0-9])"

    # This encodes the order and thing to be called for letter, we'll use it to build a tuple to pass into constructor.
    letters: dict[str, Callable] = {
        # Below 4: we will need to convert from string to int, rest of types are StrEnums so can pass string directly
        "Q": to_int,
        "P": to_int,
        "F": to_int,
        "B": to_int,
        "I": LocusProp.IllustrationType,
        "L": LocusProp.CurrierLanguage,
        "H": LocusProp.DavisHand,
        "C": LocusProp.CurrierHand,
        "X": LocusProp.ExtraneousWriting,
    }

    # Dict e.g. {'Q': 'M', 'P': 'R', ...}
    props_pairs = dict(re.findall(pattern, header))

    # For each possible property, get its value if it is present in this header, then pass it to that property's callable.
    # if it isn't present then use None.
    props = tuple(
        func(props_pairs[key]) if key in props_pairs else None
        for key, func in letters.items()
    )

    # This function is sensitive to the order of the above dict matching the order of constructor parameters. Maybe use type introspection?
    return [Locus.from_line(page_props=props, line=line) for line in lines]


def to_alphabet(text: str, alphabet: Literal["eva", "cuva"]) -> str:
    subs = [
        (r"a", "A"),
        (r"b", "B"),
        (r"cfh", "FS"),
        (r"ch", "S"),
        (r"ckh", "KS"),
        (r"cph", "PS"),
        (r"cth", "TS"),
        (r"d", "D"),
        (r"e", "E"),
        (r"ee", "U"),
        (r"eee", "UE"),
        (r"eeee", "UU"),
        (r"f", "F"),
        (r"g", "G"),
        (r"i", "I"),
        (r"ii", "N"),
        (r"iii", "M"),
        (r"iin", "M"),
        (r"iiin", "NN"),
        (r"in", "N"),
        (r"j", "Q"),
        (r"k", "K"),
        (r"l", "L"),
        (r"m", "J"),
        (r"n", "I"),
        (r"o", "O"),
        (r"p", "P"),
        (r"q", "H"),
        (r"r", "R"),
        (r"s", "C"),
        (r"sh", "Z"),
        (r"t", "T"),
        (r"u", "A"),
        (r"v", "V"),
        (r"x", "X"),
        (r"y", "Y"),
        (r"z", "J"),
    ]
    if alphabet.lower() not in ("eva", "cuva"):
        raise ValueError("alphabet must be 'eva' or 'cuva'")
    match alphabet:
        case "eva":
            return text
        case "cuva":
            for string, sub in subs:
                text = re.sub(string, sub, text)
            return text


def strip_inline_metadata(
    text: str, normalize_gaps: bool, delete_comments: bool
) -> str:
    if delete_comments:
        # There are only like 3 of these in the text
        text = re.sub(r"<[!@%$].*?>", "", text)
    if normalize_gaps:
        # Convert all <-> <~> , to .
        text = re.sub(r"<[-~]>|,", ".", text)
    return text


@dataclass
class TextProcessingOptions:
    alphabet: Literal["eva", "cuva"] = "eva"
    normalize_gaps: bool = True
    delete_comments: bool = True

    # The point of this is to funnel all the to_text, to_lines, to_words signatures to a single place,
    # by having them just pass kwargs to here
    # I really don't like that kwargs mostly destroys type checking and figuring out the signature
    # Considered decorator but it seems pretty much equivalent in effects but more complicated
    # I think this slightly edges out previous setup of having a million duplicated call signatures
    @classmethod
    def from_kwargs(cls, **kwargs):
        try:
            return cls(**kwargs)
        except TypeError as e:
            if "unexpected keyword argument" in str(e):
                param_name = str(e).split("'")[-2]
                raise ValueError(
                    f"'{param_name}' is not a valid text processing option"
                )
            else:
                raise ValueError(f"Invalid text processing options: {str(e)}")


def to_final_text(text: str, opts: TextProcessingOptions) -> str:
    text = strip_inline_metadata(
        text, normalize_gaps=opts.normalize_gaps, delete_comments=opts.delete_comments
    )
    text = to_alphabet(text, opts.alphabet)
    return text


@dataclass
class Manuscript(VMSDataclass):
    loci: list[Locus]
    source_filename: str | None = None

    @classmethod
    @lru_cache(maxsize=1)
    def from_file(cls, text: str, fname: str) -> "Manuscript":
        # Expects text directly from IVTFF file.
        # Split by page and then create Page for each.

        # Delete all comment lines
        text = re.sub(r"#.*?\n", "", text)

        loci: list[Locus] = []
        page_lines: list[str] = []

        # Going in reverse order ( [::-1] ) makes process a bit easier because we don't have to "look forward" for headings,
        # just keep piling up lines and when we do see a heading, ship it all off.
        for line in text.splitlines()[::-1]:
            # Does this line contain a page header?
            header_match = re.search(r"<! (.*?)>", line)
            if header_match:
                # Yep!
                name: str = re.match(r"<(f.*?)>", line).group(1)  # type: ignore - name is guaranteed to exist alongside header for IVTFF
                header: str = header_match.group(1)
                if len(page_lines) and name and header:
                    # Process all lines (still in reverse order)
                    page_loci = loci_list_from_lines(page_lines, header)
                    loci.extend(page_loci)
                    # Get rid of our used up lines
                    page_lines = []
            else:
                # No header in line
                page_lines.append(line)

        if len(page_lines):
            raise ValueError("File ended without final page header. Corrupted file?")

        return cls(
            loci=loci[::-1], source_filename=fname
        )  # Put loci back in normal order

    # to_lines, to_words, and the same names in VMS class are basically just wrappers for this function
    # See note in TextProcessingOptions - not super happy with kwargs
    # The below two methods should be the only methods that create TextProcessingOptions objects!
    # because they are the only ones that have to build up text in different ways by calling to_final_text
    # (it could be reduced to just to_text if Locus had some prop like last_in_page)
    def to_text(self, **kwargs) -> str:
        opts = TextProcessingOptions.from_kwargs(**kwargs)
        text = decompose(self)  # gives us a str
        text = to_final_text(text, opts)
        return text

    def to_pages(self, **kwargs) -> list[str]:
        opts = TextProcessingOptions.from_kwargs(**kwargs)
        # because of fRos, page names can't be sorted alphabetically
        pages: list[list[Locus]] = []
        curr_pagename = ""
        for locus in self.loci:
            if locus.page_name != curr_pagename:
                curr_pagename = locus.page_name
                pages.append([])
            pages[-1].append(locus)
        pagetext: list[str] = [decompose(page) for page in pages]  # type: ignore
        for i, ptext in enumerate(pagetext):
            ptext = to_final_text(ptext, opts)
            pagetext[i] = ptext
        return pagetext

    def to_lines(self, **kwargs) -> list[str]:
        return self.to_text(**kwargs).splitlines()

    def to_words(self, **kwargs) -> list[str]:
        words = []
        for line in self.to_lines(**kwargs):
            words.extend(line.split("."))
        return words


VMSObject = Manuscript | Locus


# We want to be able to easily take any combination and turn it all into text.
def decompose(source: VMSObject | list[VMSObject]) -> str:
    text = ""
    if isinstance(source, list):
        for item in source:
            text += decompose(item)
    else:
        if hasattr(source, "loci"):
            for locus in source.loci:  # type: ignore
                text += decompose(locus)
        elif hasattr(source, "text"):
            text += source.text + "\n"  # type: ignore
        else:
            raise TypeError("source must be Manuscript, Page, or Locus")
    return text


# Helper class to load from transliteration file and create Manuscript objects, returning it or its text/lines/words/filtered etc.
class VMS:
    @classmethod
    def filter(cls, props: Sequence[LocusPropType] | LocusPropType) -> Manuscript:
        if isinstance(props, LocusPropType):
            props = [props]
        filt_props = [
            prop for prop in props if issubclass(prop.__class__, LocusPropType)
        ]

        vms = cls.get()
        loci = []

        for locus in vms.loci:
            if all(fprop in locus.__dict__.values() for fprop in filt_props):
                loci.append(locus)

        if not len(loci):
            print(
                "Warning: No text loci matched the given filter properties, result is empty."
            )
            print(
                f"  (properties: {', '.join(list(f'LocusProp.{p.__class__.__name__ + "." + p.name}' for p in filt_props)) if filt_props else 'None'} )"
            )  # type: ignore

        return Manuscript(loci=loci, source_filename=vms.source_filename)

    @classmethod
    def to_words(cls, **kwargs) -> list[str]:
        return cls.get().to_words(**kwargs)

    @classmethod
    def to_lines(cls, **kwargs) -> list[str]:
        return cls.get().to_lines(**kwargs)

    @classmethod
    def to_pages(cls, **kwargs) -> list[str]:
        return cls.get().to_pages(**kwargs)

    @classmethod
    def to_text(cls, **kwargs) -> str:
        return cls.get().to_text(**kwargs)

    @classmethod
    def get(cls, basic_ver: bool = True) -> Manuscript:
        if not basic_ver:
            print(
                "Warning: Support for extended EVA very untested, please report any issues"
            )
        translit_dir = resources.files("pykeedy.data.transliterations")

        basic = translit_dir / "RF1b-er.txt"
        extended = translit_dir / "RF1b-e.txt"

        path = basic if basic_ver else extended
        with path.open("r", encoding="utf-8") as f:
            raw = f.read()

        vms = Manuscript.from_file(raw, path.name)
        return vms
